/*
 * Copyright (c) 2024, NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 */

#ifndef __VIRTNET_DPA_NET_DIM_H__
#define __VIRTNET_DPA_NET_DIM_H__

#define VIRTNET_DPA_NET_DIM_INVALID_INDEX -1

#define VIRTNET_DPA_NET_DIM_MAX_BUCKET		\
	(VIRTNET_DPA_NET_DIM_NUM_PROFILES *	\
	 VIRTNET_DPA_NET_DIM_BUCKETS_PER_PROFILE)

enum {
	/* Increasing number of DIM timeout process handlers
	 * need equivalent number of EUs.
	 * As of now, limited to 4.
	 */
	VIRTNET_DPA_NET_DIM_TIMEOUT_HANDLERS = 4,
	VIRTNET_DPA_NET_DIM_BUCKETS_PER_PROFILE = VIRTNET_DPA_NET_DIM_TIMEOUT_HANDLERS,
	VIRTNET_DPA_NET_DIM_NUM_PROFILES = 5,
	/* EU 0-4 of core 0 is available for DIM
	 * EU 0 : DIM Updater
	 * EU 1 - 4 : DIM timeout handlers
	 */
	VIRTNET_NET_DIM_MAX_EUS = 5,
	/* Number of events between DIM iterations.
	 * Causes a moderation of the algorithm run.
	 */
	VIRTNET_DPA_DIM_NEVENTS = 64,
};

struct virtnet_dpa_vq_net_dim;

/* struct virtnet_dpa_net_dim_cq_moder - Structure for CQ moderation values.
 * Used for communications between DIM and its consumer.
 *
 * @usec: CQ timer suggestion (by DIM)
 * @pkts: CQ packet counter suggestion (by DIM)
 * @cq_period_mode: CQ period count mode (from CQE/EQE)
 */
struct virtnet_dpa_net_dim_cq_moder {
	uint16_t usec;
	uint16_t pkts;
	uint8_t cq_period_mode;
};

/* struct virtnet_dpa_net_dim_sample - Structure for DIM sample data.
 * Used for communications between DIM and its consumer.
 *
 * @time: Sample timestamp
 * @pkt_ctr: Number of packets
 * @byte_ctr: Number of bytes
 * @event_ctr: Number of events
 */
struct virtnet_dpa_net_dim_sample {
	uint64_t time;
	uint32_t pkt_ctr;
	uint32_t byte_ctr;
	uint16_t event_ctr;
};

/* struct virtnet_dpa_net_dim_stats - Structure for DIM stats.
 * Used for holding current measured rates.
 *
 * @ppms: Packets per msec
 * @bpms: Bytes per msec
 * @epms: Events per msec
 */
struct virtnet_dpa_net_dim_stats {
	uint32_t ppms; /* packets per msec */
	uint32_t bpms; /* bytes per msec */
	uint32_t epms; /* events per msec */
};

/* struct virtnet_dpa_net_dim - Main structure for dynamic interrupt
 * moderation (DIM).
 * Used for holding all information about a specific DIM instance.
 *
 * @prev_stats: Measured rates from previous iteration (for comparison)
 * @start_sample: Sampled data at start of current iteration
 * @measuring_sample: A &virtnet_dpa_net_dim_sample that is used to update the
 * current events
 * @last_inspect_ticks: timestamp when we looked last for generating interrupt
 * @last_recal_ticks;	timestamp when we looked last for recalculating
 * new values
 * @state: Algorithm state
 * @profile_ix: Current moderation profile
 * @mode: CQ period count mode
 * @tune_state: Algorithm tuning state (see below)
 * @steps_right: Number of steps taken towards higher moderation
 * @steps_left: Number of steps taken towards lower moderation
 * @tired: Parking depth counter
 */
struct virtnet_dpa_net_dim {
	struct virtnet_dpa_net_dim_stats prev_stats;
	struct virtnet_dpa_net_dim_sample start_sample;
	uint64_t last_inspect_ticks;
	uint64_t last_recal_ticks;
	uint8_t state;
	uint8_t profile_ix;
	uint8_t mode;
	uint8_t tune_state;
	uint8_t steps_right;
	uint8_t steps_left;
	uint8_t tired;
};

struct virtnet_dpa_net_dim_bucket_list {
	struct spinlock_s lock;
	int16_t first;
	int16_t last;
	int16_t num;
};

struct virtnet_dpa_net_dim_entry {
	int16_t prev;
	int16_t next;
	int16_t bucket_used;
};

struct virtnet_dpa_vq_net_dim {
	struct virtnet_dpa_net_dim dim;
	struct virtnet_dpa_event_handler_ctx *event_ctx;
	struct virtnet_dpa_net_dim_entry entry;
	bool in_use;
};

/* enum virtnet_dpa_net_dim_cq_period_mode - Modes for CQ period count
 *
 * @VIRTNET_DPA_NET_DIM_CQ_PERIOD_MODE_CQE: Start counting from CQE
 * @VIRTNET_DPA_NET_DIM_CQ_PERIOD_NUM_MODES: Number of modes
 */
enum virtnet_dpa_net_dim_cq_period_mode {
	VIRTNET_DPA_NET_DIM_CQ_PERIOD_MODE_CQE = 0x0,
	VIRTNET_DPA_NET_DIM_CQ_PERIOD_NUM_MODES
};

/* enum virtnet_dpa_net_dim_state - DIM algorithm states
 *
 * These will determine if the algorithm is in a valid state to start an
 * iteration.
 *
 * @VIRTNET_DPA_NET_DIM_START_MEASURE: This is the first iteration
 * (also after applying a new profile)
 * @VIRTNET_DPA_NET_DIM_MEASURE_IN_PROGRESS: Algorithm is already in progress -
 * check if need to perform an action
 * @VIRTNET_DPA_NET_DIM_APPLY_NEW_PROFILE: DIM consumer is currently applying a
 * profile - no need to measure
 */
enum virtnet_dpa_net_dim_state {
	VIRTNET_DPA_NET_DIM_START_MEASURE,
	VIRTNET_DPA_NET_DIM_MEASURE_IN_PROGRESS,
	VIRTNET_DPA_NET_DIM_APPLY_NEW_PROFILE,
};

/* enum virtnet_dpa_net_dim_tune_state - DIM algorithm tune states
 *
 * These will determine which action the algorithm should perform.
 *
 * @VIRTNET_DPA_NET_DIM_PARKING_ON_TOP: Algorithm found a local top point -
 * exit on significant difference
 * @VIRTNET_DPA_NET_DIM_PARKING_TIRED: Algorithm found a deep top point -
 * don't exit if tired > 0
 * @VIRTNET_DPA_NET_DIM_GOING_RIGHT: Algorithm is currently trying higher
 * moderation levels
 * @VIRTNET_DPA_NET_DIM_GOING_LEFT: Algorithm is currently trying lower
 * moderation levels
 */
enum virtnet_dpa_net_dim_tune_state {
	VIRTNET_DPA_NET_DIM_PARKING_ON_TOP,
	VIRTNET_DPA_NET_DIM_PARKING_TIRED,
	VIRTNET_DPA_NET_DIM_GOING_RIGHT,
	VIRTNET_DPA_NET_DIM_GOING_LEFT,
};

/* enum virtnet_dpa_dim_stats_state - DIM algorithm statistics states
 *
 * These will determine the verdict of current iteration.
 *
 * @VIRTNET_DPA_DIM_STATS_WORSE: Current iteration shows worse performance
 * than before
 * @VIRTNET_DPA_DIM_STATS_SAME:  Current iteration shows same performance
 * than before
 * @VIRTNET_DPA_DIM_STATS_BETTER: Current iteration shows better performance
 * than before
 */
enum virtnet_dpa_dim_stats_state {
	VIRTNET_DPA_DIM_STATS_WORSE,
	VIRTNET_DPA_DIM_STATS_SAME,
	VIRTNET_DPA_DIM_STATS_BETTER,
};

/* enum virtnet_dpa_net_dim_step_result - DIM algorithm step results
 *
 * These describe the result of a step.
 *
 * @VIRTNET_DPA_NET_DIM_STEPPED: Performed a regular step
 * @VIRTNET_DPA_NET_DIM_TOO_TIRED: Same kind of step was done multiple times -
 * should go to tired parking
 * @VIRTNET_DPA_NET_DIM_ON_EDGE: Stepped to the most left/right profile
 */
enum virtnet_dpa_net_dim_step_result {
	VIRTNET_DPA_NET_DIM_STEPPED,
	VIRTNET_DPA_NET_DIM_TOO_TIRED,
	VIRTNET_DPA_NET_DIM_ON_EDGE,
};

/* virtnet_dpa_net_dim_update_sample - set a sample's fields with given values
 * @s: DIM sample
 * @event_ctr: number of events to set
 * @packets: number of packets to set
 * @bytes: number of bytes to set
 */
static inline void
virtnet_dpa_net_dim_update_sample(struct virtnet_dpa_net_dim_sample *s,
				  uint16_t event_ctr, uint64_t packets,
				  uint64_t bytes, uint64_t time)
{
	s->time	= time;
	s->pkt_ctr = packets;
	s->byte_ctr = bytes;
	s->event_ctr = event_ctr;
}

/* virtnet_dpa_net_dim_get_rx_moderation - provide a CQ moderation object for
 * the given RX profile
 * @cq_period_mode: CQ period mode
 * @ix: Profile index
 */
struct virtnet_dpa_net_dim_cq_moder *
virtnet_dpa_net_dim_get_rx_moderation(uint8_t cq_period_mode, int ix);

/* virtnet_dpa_net_dim_get_tx_moderation - provide a CQ moderation object for
 * the given TX profile
 * @cq_period_mode: CQ period mode
 * @ix: Profile index
 */
struct virtnet_dpa_net_dim_cq_moder *
virtnet_dpa_net_dim_get_tx_moderation(uint8_t cq_period_mode, int ix);

/* virtnet_dpa_net_dim - main DIM algorithm entry point
 * @dim: DIM instance information
 * @end_sample: Current data measurement
 *
 * Called by the consumer.
 * This is the main logic of the algorithm, where data is processed in order
 * to decide on next required action.
 */
void virtnet_dpa_net_dim_run(struct virtnet_dpa_net_dim *dim,
			     struct virtnet_dpa_net_dim_sample end_sample);

void
virtnet_dpa_net_dim_event_ctx_sample_update(struct virtnet_dpa_event_handler_ctx *event_ctx,
					uint32_t pkt_count, uint32_t byte_count,
					uint16_t event_count);
void
virtnet_dpa_net_dim_nw_handler_msix_send(struct virtnet_dpa_event_handler_ctx *event_ctx,
					 uint32_t iter_num_cqe,
					 uint32_t total_bytes);
int
virtnet_dpa_net_dim_config(void *dev, void *ibv_ctx, bool net_dim_enable);
#endif /* __VIRTNET_DPA_NET_DIM_H__ */
