/*
 * Copyright (c) 2023, NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 */

#ifndef __VIRTNET_DPA_TX_SCHED_COMMON_H__
#define __VIRTNET_DPA_TX_SCHED_COMMON_H__

enum virtnet_dpa_tx_sched_pool_sz {
	/* The maximum number of cores for virtnet is 0-10, the rest is for PCC
	 * and RTOS. EU 5 is assigned for schedule handler and poll work
	 * handler. Schedule handler uses core[0]EU[5]. Poll work handler uses
	 * core[1-5]EU[5]. But if number of cores for virtnet is 0-3,
	 * poll pool size is 3, schedule handler uses core[0]EU[5], poll work
	 * handler uses core[1-3]EU[5].
	 */
	TX_SCHED_POLL_HDLR_POOL_MAX_SZ = 5,
	/* Event handler uses core[0-10]EU[6-10]. It is 55 EUs in total, the
	 * size of event handler should be double of assigned EU number in
	 * statistically.
	 */
	TX_SCHED_EVENT_HDLR_POOL_SZ = 55,
};

enum virtnet_dpa_tx_sched_rpc_type {
	VIRTNET_DPA_TX_SCHED_RPC_HDLR_INIT = 0,
	VIRTNET_DPA_TX_SCHED_RPC_HDLR_ABORT = 1,
};

/* Save common info for tx/rx, rx scheduler will be support later */
struct virtnet_dpa_sched_hdlr_info {
	/* Use this cqn to wake up handler thread */
	struct virtnet_dpa_cq_ctx trigger_hdlr_cq_ctx;
	uint32_t emu_mng_gvmi;
	/* Stop handler by RPC */
	uint8_t abort_flag;
	uint16_t eu_num;
};

/* Common for tx/rx, rx scheduler will be support later */
struct virtnet_dpa_sched_dev_hash_entry {
	uint32_t emu_db_to_cq_id;
	uint16_t seed_flag;
	uint16_t event_ctx_id;
};

/* Common handler context for tx/rx, rx scheduler will be support later.
 * work_entry is shared memory for scheduler and poll/event handler. Scheduler
 * uses emu_db_handle to found hash_entry which saves event_ctx_id and writes
 * hash_entry address in work_entry. Poll/event handler reads work_entry and
 * counts event_ctx address, then calls virtnet_sq_db_handler to process.
 */
struct virtnet_dpa_sched_work_entry {
	struct virtnet_dpa_sched_dev_hash_entry *hash_entry;
};

struct virtnet_dpa_tx_sched_counters {
	uint64_t sched_hdlr;
	uint64_t poll_hdlr;
	uint64_t event_hdlr;
	uint64_t error_cqe;
	uint64_t hash_table_miss;
};

/* Common handler context for tx/rx poll handler, rx scheduler will be support
 * later
 */
struct virtnet_dpa_sched_poll_hdlr_ctx {
	struct virtnet_dpa_sched_hdlr_info hdlr_info;
	struct virtnet_dpa_tx_sched_poll_work *work;
	flexio_uintptr_t event_ctx_pool;
};

struct virtnet_dpa_tx_sched_poll_hdlr_ctx {
	struct virtnet_dpa_sched_poll_hdlr_ctx ctx;
};

/* Each poll handler has one work_entry */
struct virtnet_dpa_tx_sched_poll_work {
	struct virtnet_dpa_sched_work_entry work_entry;
};

struct virtnet_dpa_tx_sched_poll_rpc_ctx {
	enum virtnet_dpa_tx_sched_rpc_type rpc_type;
	/* rpc triggers all poll handlers. Points to the first poll
	 * handler ctx
	 */
	struct virtnet_dpa_tx_sched_poll_hdlr_ctx *hdlr_ctx;
	uint8_t hdlr_num;
};

struct virtnet_dpa_tx_sched_event_wq{
	/* Write by event worker and read by scheduler */
	uint32_t ci;
	/* write by scheduler and read by event worker */
	uint32_t pi;
	/* Each event work has a queue can hold 2^3 sq to process. */
	#define TX_SCHED_EVENT_WQ_LOG_DEPTH 3
	struct virtnet_dpa_sched_work_entry
				work_entry[POW2(TX_SCHED_EVENT_WQ_LOG_DEPTH)];
};

struct virtnet_dpa_sched_event_hdlr_ctx {
	struct virtnet_dpa_sched_hdlr_info hdlr_info;
	struct virtnet_dpa_tx_sched_event_wq *wq;
	flexio_uintptr_t event_ctx_pool;
};

struct virtnet_dpa_tx_sched_event_hdlr_ctx {
	struct virtnet_dpa_sched_event_hdlr_ctx ctx;
};

struct virtnet_dpa_tx_sched_event_rpc_ctx {
	enum virtnet_dpa_tx_sched_rpc_type rpc_type;
	/* rpc triggers all event handlers. This points to the first device
	 * event handler ctx
	 */
	struct virtnet_dpa_tx_sched_event_hdlr_ctx *hdlr_ctx;
};

/* Tx scheduler handler context */
struct virtnet_dpa_tx_sched_hdlr_ctx {
	/* hash key is organized with virtnet_dpa_tx_sched_hash_key()
	 * hash entry is organized with virtnet_dpa_tx_sched_hash_data()
	 */
	struct virtnet_dpa_sched_dev_hash_entry
					hash_entry[TX_SCHED_HASH_TBL_SIZE];
	struct virtnet_dpa_sched_hdlr_info hdlr_info;
	/* Scheduler handler allocates work_entry(shared memory), poll handler
	 * saves pointer
	 */
	struct virtnet_dpa_tx_sched_poll_work
				poll_work[TX_SCHED_POLL_HDLR_POOL_MAX_SZ];
	/* Scheduler handler allocates shared memory, event handler saves shared
	 * memory point
	 */
	struct virtnet_dpa_tx_sched_event_wq
					event_wq[TX_SCHED_EVENT_HDLR_POOL_SZ];
	/* Scheduler handler uses below info to trigger event handlers up */
	struct virtnet_dpa_tx_sched_event_hdlr_ctx *event_hdlr_ctx;
	/* schedule handler counters */
	struct virtnet_dpa_tx_sched_counters sched_counters;
	/* Start index for event work round robin queue. */
	uint32_t start_hdlr;
	uint8_t poll_hdlr_num;
};

/* Tx scheduler RPC context */
struct virtnet_dpa_tx_sched_rpc_ctx {
	enum virtnet_dpa_tx_sched_rpc_type rpc_type;
	struct virtnet_dpa_tx_sched_hdlr_ctx *hdlr_ctx;
};

/* For host to query tx scheduler data */
struct virtnet_dpa_tx_sched_data {
	struct virtnet_dpa_tx_sched_hdlr_ctx hdlr_ctx;
};

static inline uint32_t
virtnet_dpa_tx_sched_hash_key(uint32_t emu_db_to_cq_id)
{
	/* As sq always exhibits the same parity, the last digit of cq_id is
	 * removed to calculate the hash key. This helps achieve a better
	 * distribution of entries.
	 */
	return (emu_db_to_cq_id >> 1) % TX_SCHED_HASH_TBL_SIZE;
}

/* Use linear probing to get hash entry
 * emu_db_to_cq_id: compare emu_db_to_cq_id to get the correct entry
 * VIRTNET_DPA_EHCTX_DADDR_HASH_SEED: this entry has data
 * event_ctx_id: get the address of event handler context
 */
static inline void
virtnet_dpa_tx_sched_hash_data(uint32_t emu_db_to_cq_id, uint16_t event_ctx_id,
			       struct virtnet_dpa_sched_dev_hash_entry *entry)
{
#define VIRTNET_DPA_EHCTX_DADDR_HASH_SEED 0x00000001
	entry->emu_db_to_cq_id = emu_db_to_cq_id;
	entry->seed_flag = VIRTNET_DPA_EHCTX_DADDR_HASH_SEED;
	entry->event_ctx_id = event_ctx_id;
}
#endif
